/*
 * $RCSfile: sscanf.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: sscanf.c,v $
 * Revision 1.1.1.1  1994/02/09 22:48:52  menze
 * Moved contents of _doscan.c into this file
 *
 * Revision 1.1  1994/01/31  01:07:44  menze
 * Initial revision
 *
 * ------------------------------------------------------------
 * David Mosberger's sscanf implementation.
 */


/*
 * Subset of ANSI C scanf() scanner.  The following things are
 * currently unsupported:
 *
 *	- floating point conversions
 *	- %n format (to return # of characters read so far)
 *	- scan sets
 *	- field width's
 *
 * Also, the first unmatched character in the input string is
 * consumed.  This is incompatible with the standard definition of
 * scanf.
 */
#include "platform.h"

#define FMT_SHORT	(1<<0)	/* short value? */
#define FMT_LONG	(1<<1)	/* long value? */
#define FMT_SIGNED	(1<<2)	/* signed value? */
#define FMT_SUPPRESS	(1<<3)	/* suppress assignment? */
#define FMT_BASED	(1<<4)	/* allow leading 0 or 0x to indicate base? */


/*
 * Macro to check whether "ch" is white-space.  Careful: "ch"
 * is evaluated multiple times!
 */
#define	WS(ch)	((ch) == ' ' || (ch) == '\t' || (ch) == '\n')


/*
 * Return the digit represented by character "ch" in base "base" or
 * -1 if it's not a valid digit in that base.  Upper and lower-case
 * characters are both allowed.
 */
static int
digit(int ch, int base)
{
    if (base <= 10) {
	if (ch >= '0' && ch < '0' + base) {
	    return ch - '0';
	} /* if */
    } else {
	if (ch >= '0' && ch <= '9') {
	    return ch - '0';
	} else if (ch >= 'a' && ch < 'a' + base - 10) {
	    return ch - 'a' + 10;
	} else if (ch >= 'A' && ch < 'A' + base - 10) {
	    return ch - 'A' + 10;
	} /* if */
    } /* if */
    return -1;
} /* digit */


static int
scan_int(va_list *ap, int *ch2, int base, int flags,
	 int (*inc)(void *), void *arg,
	 int *countp)
{
    int negative = 0;
    int d;
    u_long v;
    void *p;

    if (flags & FMT_SIGNED) {
	if (*ch2 == '-') {
	    negative = 1; *ch2 = (*inc)(arg);
	} else if (*ch2 == '+') {
	    *ch2 = (*inc)(arg);
	} /* if */
    } /* if */

    if ((flags & FMT_BASED) && (*ch2 == '0')) {
	*ch2 = (*inc)(arg);
	base = 8;
	if (*ch2 == 'x' || *ch2 == 'X') {
	    *ch2 = (*inc)(arg);
	    base = 16;
	} /* if */
    } /* if */

    v = 0;
    d = digit(*ch2, base);
    if (d < 0) {
	return FALSE;		/* mismatch */
    } /* if */
    do {
	v = base * v + d; *ch2 = (*inc)(arg);
	d = digit(*ch2, base);
    } while (d >= 0);
    if (negative) {
	v = -(long int) v;
    } /* if */

    if (!(flags & FMT_SUPPRESS)) {
	++(*countp);
	p = va_arg(*ap, void *);
	if (flags & FMT_SHORT) {
	    *(u_short*) p = v;
	} else if (flags & FMT_LONG) {
	    *(u_long*) p = v;
	} else {
	    *(u_int*) p = v;
	} /* if */
    } /* if */
    return TRUE;
} /* scan_int */


/* 
 * This _doscan can be made global to support scanf, fscanf, etc.
 * Since we only implement sscanf here, it's static.
 */
static int
_doscan(const char *format, va_list *ap, int (*inc)(void *), void *arg)
{
    int ch1, ch2, base;
    int count = 0;
    int flags;
    int width;
    u_long v;
    u_char *cp;

    ch1 = *format++;
    ch2 = (*inc)(arg);
    while (ch1 && (ch2 > 0)) {
	if (WS(ch1)) {
	    ch1 = *format++;
	    /* skip white space: */
	    while (WS(ch2)) {
		ch2 = (*inc)(arg);
	    } /* while */
	} else if (ch1 == '%') {
	    ch1 = *format++;
	    flags = 0;
	    v = 0;

	    /* check for assignment suppression: */
	    if (ch1 == '*') {
		ch1 = *format++;
		flags |= FMT_SUPPRESS;
	    } /* if */

	    /* scan field width: */
	    width = 0;
	    while (ch1 >= '0' && ch1 <= '9') {
		width = 10 * width + (ch1 - '0'); ch1 = *format++;
	    } /* if */
	    if (width > 0) {
		printf(__FILE__":_doscan: width specs unsupported\n");
	    } /* if */

	    /* scan argument size: */
	    if (ch1 == 'h') {
		flags |= FMT_SHORT; ch1 = *format++;
	    } else if (ch1 == 'l') {
		flags |= FMT_LONG; ch1 = *format++;
	    } /* if */

	    /* skip white space (except for 'c' and '['): */
	    if (ch1 != 'c' && ch1 != '[') {
		while (WS(ch2)) {
		    ch2 = (*inc)(arg);
		} /* while */
	    } /* if */

	    base = 10;
	    switch (ch1) {
	      case '%':
		if (ch2 != '%') {
		    return count;
		} /* if */
		ch2 = (*inc)(arg);
		break;

	      case 'c':
		if (!(flags & FMT_SUPPRESS)) {
		    ++count;
		    cp = va_arg(*ap, u_char *);
		    *cp = ch2;
		} /* if */
		ch2 = (*inc)(arg);
		break;

	      case 's':
		if (!(flags & FMT_SUPPRESS)) {
		    ++count;
		    cp = va_arg(*ap, u_char *);
		    while (!WS(ch2)) {
			*cp++ = ch2; ch2 = (*inc)(arg);
		    } /* while */
		    *cp = '\0';
		} /* if */
		break;

	      case 'i':
		flags |= FMT_BASED;
	      case 'd':
		flags |= FMT_SIGNED;
	      case 'u':
		if (!scan_int(ap, &ch2, base, flags, inc, arg, &count)) {
		    return count;
		} /* if */
		break;
		
	      case 'o':
		if (!scan_int(ap, &ch2, 8, flags, inc, arg, &count)) {
		    return count;
		} /* if */
		break;

	      case 'x': case 'X':
	      case 'p':
		if (!scan_int(ap, &ch2, 16, flags, inc, arg, &count)) {
		    return count;
		} /* if */
		break;

	      default:
		/* unknown/unsupported format: force premature end */
		printf(__FILE__":_doscan: format `%%%c' unsupported\n", ch1);
		return count;
	    } /* switch */
	    ch1 = *format++;
	} else {
	    if (ch1 != ch2) {
		/* mismatch */
		return count;
	    } /* if */
	    ch1 = *format++;
	    ch2 = (*inc)(arg);
	} /* if */
    } /* while */
    if (ch2 < 0) {
	return ch2;
    } /* if */
    return count;
} /* _doscan */



static int
sgetchar(void *arg)
{
    char **string = (char**)arg;
    int ch;

    ch = **string;
    ++(*string);
    return ch;
} /* sgetchar */


#if 0
    /* 
     * The right way ...
     */
int
sscanf(const char *string, const char *format, ...)
#else
int
sscanf( string, format, va_alist )
    const char *string;
    const char *format;
    va_dcl
#endif

{
    va_list ap;
    int count;

#if 0
    /* 
     * The right way ...
     */
    va_start(ap, format);
#else
    va_start(ap);
#endif
    count = _doscan(format, &ap, sgetchar, &string);
    va_end(ap);
    return count;
} /* sscanf */

			/*** end of sscanf.c ***/
