/*
 * $RCSfile: rsasign.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 *
 * $Log: rsasign.c,v $
 * Revision 1.3  1996/05/28 18:02:57  ho
 * prototype error; bogus use of null pointer.
 *
 * Revision 1.2  1996/05/20  22:19:16  rajesh
 * Added a rom option and made changes for sending the source
 * address with the message, so that it could be used to
 * verify the signature.
 *
 * Revision 1.7  1996/04/06  02:43:29  rcs
 * swapped which key was used to send versus receive, so that remote key
 * is now used to send.  Changed various constants designating "remote"
 * etc. to #defined values in _i.h.  Removed useless code for ivec.
 * Fixed bug in rsaCreateSessn that was only allocating two slots for
 * SessnArray on stack.  Corrected some comments.  Made Open check that
 * KM exists (as well as transport).
 *
 * Revision 1.6  1994/02/05  00:05:02  menze
 *   [ 1994/02/03          menze ]
 *   Added casts
 *
 * Revision 1.5  1994/01/08  01:44:46  rcs
 * Removed most dead code, and PRESERVE_LENGTH flag leftover from CRYPT.
 * Cleaned up comments to be more accurate.
 * Removed a bunch of #if 0 code.
 * Didn't change anything that actually executes.
 *
 * Revision 1.4  1993/12/12  05:06:26  menze
 * Fixed a comment-out that conflicted with endif comments
 *
 */


/*  CRYPTOGRAPHIC SHORTCOMINGS

This version of RSA is experimental in various ways, so
some cryptographic Is and Ts remain undotted and uncrossed.
It doesn't zero out message plaintext after encryption, or
ciphertext after decryption.  Keys aren't zeroed out when the
sessions are closed.  There's a small storage leak when a session
is closed because the key memory isn't returned.
This version of RSA operates in ECB (electronic code book) mode, so,
in a message that contains a repeated block of plaintext, the plaintext
will be enciphered to the same value each time.
A block of 0s enciphers to a block of 0s, and a block with
all zeros except the low bit will encipher to the same thing.
No effort has been made to zero out GNU intermediate variables.
The GNU bignum routines are still using regular malloc & free;
this should be changed to the xkernel "secure" versions.
RSA is quite slow.
There is no inter-message cryptographic chaining.
Our current system of one key per host is suitable for testing,
but wouldn't be appropriate for a real application.  We haven't
taken any special precautions to protect the file of keys.

The information passed during opening of a session includes
fairly arbitrary participants; this provides an outbound information
channel for Trojan Horse client programs.  The lengths of outgoing
packets are another channel.  The presence or absence of communication
is a third channel.  A Trojan Horse client can also originate known
plaintext to help out a remote attacker.

Moving to the wire,
the source & destination IP addresses appear in clear for every
packet on the net, so traffic analysis is trivial for a properly
located eavesdropper.  Message length is also available to the
eavesdropper.  Many protocols have standard length messages they
send (acknowledgements, routing updates, etc.), often with fixed,
known, or easily guessable contents.  These are also available to
an opponent.  Many hardware platforms are capable of lying about
packet origination address.  The Internet routing protocols are
presently unprotected, allowing for various kinds of mischief.

*/


#include "xkernel.h"
#include "rsasign.h"
#include "rsasign_i.h"
#include "gmp.h"
#include "eth.h"		/* So ARP can go on top of RSA */
#include "ip.h"                /* For IP pseudoheader fixup. */
#include "hash.h"
#include "km.h"

extern int tracersasignp; int tracersasignp;

/* Address alignment note:  The data buffer addresses must be word aligned (addresses
   a multiple of 4 bytes, unless the machine architecture will handle load/store of
   words on non-aligned addresses.  Presently, the crypt routines always allocate
   their own working arrays, so the alignment is guaranteed by xMalloc, etc.
*/

/* No testing on Little-Endian machines yet.  No byte rearrangement is done.
On Little-Endian machines, a byte-swap+unswap is probably needed around each
en/decryption, if we want to communicate between machines of opposite endianness.
*/

/* key: modulus & exponents for testing
   primes checked with quick-prime; not fully proved

   p1 3141590000000000000000000000000000000000000000000000000000000000000000000143
      first prime after 314159 * 10^70
   p2 4267841421045584274069507796436150524886877606750846984400752944757834099759331
      largest prime < 2^512/p1
   mod = p1 * p2
   phi = (p1-1) * (p2-1)
   encryption exponent = 2^511 + 3
   decryption exponent = 1/enc (mod phi)
*/

/*
char *myrsakeymod = "13407807929942597099574024998205846127479365820592393377723561443721764030073217999499518551191939614890369525058823497765371118769307671100370276265584333";
char *myrsakeyenc = "6703903964971298549787012499102923063739682910296196688861780721860882015036773488400937149083451713845015929093243025426876941405973284973216824503042051";
char *myrsakeydec = "2210719290388045346905339355787434075468743397471703259900205710040069873528157578187197280387011711560353471825220112394617831579775760441554445681825271";
MP_INT rsakeymod,rsakeyenc,rsakeydec;
*/
/* long rsakeysize = 64;*/ /* size in bytes */
/* long rsalongsize;*/     /* size in longs, rounded up */
/* note on the keysize-in-bytes:  We promise that keysize bytes is enough to
   hold any residue mod M, and that keysize-1 bytes can always be represented
   by some residue mod M.  If M were an exact power of 2 (highly unrecommended),
   these two lengths could be the same number.  I have selected up-rounding,
   so for example, M=2^32 will have a 5 byte keysize. */

/* An RSA key is a set of three numbers, separated by any nonnumeric characters.
   The three numbers are the modulus, the encryption exponent, and the
   decryption exponent, in decimal.  The decryption exponent may be 0 if absent.
   A bad format key can cause a core dump in the key parser.
   No consistency checking is done on the numbers.
   No cryptographic care is taken to erase intermediate results.   */
/*typedef struct{ MP_INT modulus,encexp,decexp; long keysize;} rsa_keyblock; */

/*$$$$ contains bare printfs; needs failure modes */
static rsa_keyblock* rsa_parse_key(kstr) char *kstr;
{  rsa_keyblock *kblk = (rsa_keyblock *) malloc(sizeof(rsa_keyblock)); /* use xmalloc $$$$ */
   char *kp = kstr, ktmp[2001]; int len;

#ifdef DEBUG
   printf("rsa_parse_key: arg = |%s|\n",kstr);
#endif
   kp += strcspn(kp,"0123456789");
   len = strspn(kp,"0123456789");
   if (len>2000) printf("rsa_parse_key:  bad number\n");
   strncpy(ktmp,kp,len); ktmp[len]='\0'; kp += len;
   mpz_init_set_str(&kblk->modulus,ktmp,10);
#ifdef DEBUG
   printf("modulus = "); mpz_out_str(stdout,10,&kblk->modulus); printf("\n");
#endif
   /* mpz_sizeinbase gives 1 + floor(log2 M) */
   kblk->keysize = (mpz_sizeinbase(&kblk->modulus,2) + 7) / 8;
#ifdef DEBUG
   printf("keysize = %d\n", kblk->keysize);
#endif

   kp += strcspn(kp,"0123456789");
   len = strspn(kp,"0123456789");
   if (len>2000) printf("rsa_parse_key:  bad number\n");
   strncpy(ktmp,kp,len); ktmp[len]='\0'; kp += len;
   mpz_init_set_str(&kblk->encexp,ktmp,10);
#ifdef DEBUG
   printf("encexp = "); mpz_out_str(stdout,10,&kblk->encexp); printf("\n");
#endif

   kp += strcspn(kp,"0123456789");
   len = strspn(kp,"0123456789");
   if (len>2000) printf("rsa_parse_key:  bad number\n");
   strncpy(ktmp,kp,len); ktmp[len]='\0'; kp += len;
   mpz_init_set_str(&kblk->decexp,ktmp,10);
#ifdef DEBUG
   printf("decexp = "); mpz_out_str(stdout,10,&kblk->decexp); printf("\n");
#endif

   /* save the parsed string, so we can compare it later;
      we save the whole string, even though we didn't look at anything past
      the first three numbers */
   kblk->key_string = (u_char*)xMalloc(strlen(kstr)+1);
   strcpy((char *)kblk->key_string,kstr);
   
   return kblk;
 }

/* look more closely into exactly which storage this returns;
   it should probably be zeroed out before giving it back;
   I might change x-mp-free to do zeroing $$$$ */
static void zap_MP_INT(mpint) MP_INT* mpint;
{  mpz_clear(mpint);  }

/* static the routines $$$ */
static void rsa_zap_keyblock(kblk) rsa_keyblock *kblk;
{  if (kblk->key_string) { bzero((char*)kblk->key_string,strlen((char *)kblk->key_string));
			   xFree((char*)kblk->key_string); };
   if (kblk->modulus.alloc) zap_MP_INT(&kblk->modulus);
   if (kblk->encexp.alloc) zap_MP_INT(&kblk->encexp);
   if (kblk->decexp.alloc) zap_MP_INT(&kblk->decexp);
   bzero((char*)kblk,sizeof(rsa_keyblock));  }


#ifdef __STDC__

static XObj		rsaCreateSessn( XObj, XObj, XObj, XObj, Part *);
static void		rsaSessnInit( XObj );
static int		rsaControlProtl( XObj, int, char *, int );
static xkern_return_t	rsaDemux( XObj, XObj, Msg * );
static XObj		rsaOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	rsaOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	rsaOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	rsaOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	rsaClose( XObj );
static xmsg_handle_t	rsaPush( XObj, Msg * );
static xkern_return_t	rsaPop( XObj, XObj, Msg *, void * );
static int		rsaControlSessn( XObj, int, char *, int );
static void		rsaStore( VOID *, char *, long, VOID * );
static long		rsacopy( VOID *, char *, long, VOID * );
static long             chkPop(VOID *, char *, long, void *);
static xkern_return_t 	romsendsource(XObj, char **, int, int, void *);

#else

static XObj		rsaCreateSessn();
static void		rsaSessnInit();
static int		rsaControlProtl();
static xkern_return_t	rsaDemux();
static XObj		rsaOpen();
static xkern_return_t	rsaOpenEnable();
static xkern_return_t	rsaOpenDisable();
static xkern_return_t	rsaOpenDone();
static xkern_return_t	rsaClose();
static xmsg_handle_t	rsaPush();
static xkern_return_t	rsaPop();
static int		rsaControlSessn();
static xkern_return_t 	romsendsource();

#endif /* __STDC__ */

static XObjRomOpt rmopts[] = {
                               { "sendsource", 2, romsendsource },
			     };

void
rsasign_init( self )
     XObj self;
{
    PState	*ps;

    xTrace0(rsasignp, TR_ERRORS, "rsa_init entered");
    xAssert(xIsProtocol(self));
    
    ps = X_NEW(PState);
    self->state = (VOID *)ps;

    if (self->numdown >= 3)
    	ps->leafmode = 0;
    else
	ps->leafmode = 1;

    if (ps->leafmode) {
        xTrace0(rsasignp, TR_MAJOR_EVENTS, "rsa_init : in leaf mode");
	if ( !(xIsProtocol(xGetDown(self, LEAF_RSA_PROT_CHECKSUM))
	    && xIsProtocol(xGetDown(self, LEAF_RSA_PROT_KMGR))  ) ) { 
	    xError("rsa_init: rsa cannot get lower protocol object");	
	    return; 
	}
    } else if ( !(xIsProtocol(xGetDown(self, RSA_PROT_TRANSP))
	    && xIsProtocol(xGetDown(self, RSA_PROT_CHECKSUM))
	    && xIsProtocol(xGetDown(self, RSA_PROT_KMGR))  )) { 
	xError("rsa_init: rsa cannot get lower protocol object");	
	return; 
    }

    ps->actMap = mapCreate(RSA_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(RSA_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->flags = 0;  /* Only flag defined is &1, turning on IPpseudoheader length fixup. */

    self->control = rsaControlProtl;
    self->open = rsaOpen;
    self->openenable = rsaOpenEnable;
    self->opendisable = rsaOpenDisable;
    self->opendone = rsaOpenDone;
    self->close = rsaClose;  
    self->demux = rsaDemux;

    ps->sendsource = 0;
    findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(XObjRomOpt), 0);

    xTrace0(rsasignp, TR_EVENTS, "rsa_init exiting");
}


static XObj
rsaOpen( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    XObj 	rsaSessn;
    XObj	lls;
    PState	*pstate;

    pstate = (PState *) self->state;
    
    if ( part == 0 || partLen(part) < 1 || partLen(part) > 2 ) {
	xTrace0(rsasignp, TR_SOFT_ERRORS, "rsaOpen -- bad participants");
	return ERR_XOBJ;
    }

    xTrace4(rsasignp,TR_EVENTS,"rsaOpen calling xOpen with args %x %x %x %x",
	    self,hlpType,xGetDown(self,RSA_PROT_TRANSP),part);

    if (!pstate->leafmode) {
	lls = xOpen(self, hlpType, xGetDown(self, RSA_PROT_TRANSP), part);
    	if ( lls == ERR_XOBJ ) {
             xTrace0(rsasignp, TR_ERRORS, "rsaOpen: could not open lls");
             return ERR_XOBJ;
    	}

    	if ( mapResolve(pstate->actMap, &lls, &rsaSessn) == XK_SUCCESS ) {
            xTrace0(rsasignp, TR_MAJOR_EVENTS, "found an existing one");
            xClose(lls);
    	} else {
            xTrace0(rsasignp, TR_MAJOR_EVENTS, "creating a new one");
            if ( (rsaSessn = rsaCreateSessn(self, hlpRcv, hlpType, lls, part)) == ERR_XOBJ ) {
            	xClose(lls);
            }
    	}
    }
    else {
        if ( (rsaSessn = rsaCreateSessn(self, hlpRcv, hlpType, lls, part)) == ERR_XOBJ ) {
	    xTrace0(rsasignp, TR_ERRORS, "rsaOpen: create session failed");
        }
    }

    return (rsaSessn);
}


static XObj
rsaCreateSessn( self, hlpRcv, hlpType, lls, p )
    XObj	self, hlpRcv, hlpType, lls; 
    Part 	*p;
{
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    Part        part[2];
    char        partBuf[2000];
    XObj        SessnArray[4];
    XObj        kmsrem;
    XObj        kmsloc;
    XObj        chks;
    Part	partrem[5];
    Part	partloc[5];
    int		kmpi, chkpi;

    xTrace4(rsasignp, TR_EVENTS, "rsaCreateSessn entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, lls);

    ss = X_NEW(SState);

    
    if (ps->leafmode) {
	chkpi = LEAF_RSA_PROT_CHECKSUM;
	kmpi = LEAF_RSA_PROT_KMGR;
    }
    else {
	chkpi = RSA_PROT_CHECKSUM;
	kmpi = RSA_PROT_KMGR;
    }

    if (ps->leafmode) {
	bcopy (p, part, sizeof(part));
    }
    else {
	if (xControl(lls, GETPARTICIPANTS, partBuf, 100) <= 0) {
	    xTrace0(rsasignp, TR_ERRORS, "rsaCreateSessn: could not get participants");
	    xFree((char *) ss);
	    return ERR_XOBJ;
	}
	partInternalize(part,partBuf);
    }

    if (partLen(part) < 2) {
	xTrace0(rsasignp, TR_ERRORS, "rsaCreateSessn: too few participants");
	xFree((char *) ss);
	return ERR_XOBJ;
    }

    /* open the checksum session */
    chks = xOpen(self, self, xGetDown(self, chkpi), p);
    if (chks == ERR_XOBJ) { 
	xTrace0(rsasignp, TR_ERRORS, "rsaCreateSessn: could not open checksum session");
	xFree((char *) ss);
	return ERR_XOBJ;
    }

    /* open two key manager sessions */
    partInit(partrem,1);
    partrem[0] = part[RSA_PART_REMOTE];
    partrem[0].len =1; /* major hack */
    kmsrem = xOpen(self, self, xGetDown(self, kmpi), partrem);
    if (kmsrem == ERR_XOBJ) { 
	xTrace0(rsasignp, TR_ERRORS, "rsaCreateSessn: could not open key manager");
	xFree((char *) ss);
	return ERR_XOBJ;
    }

    partInit(partloc,1);
    partloc[0] = part[RSA_PART_LOCAL];
    partloc[0].len =1; /* major hack */
    kmsloc = xOpen(self, self, xGetDown(self, kmpi), partloc);
    if (kmsloc == ERR_XOBJ) { 
	xTrace0(rsasignp, TR_ERRORS, "rsaCreateSessn: could not open key manager");
	xClose(kmsrem);
	xFree((char *) ss);
	return ERR_XOBJ;
    }


    SessnArray[RSA_SESS_TRANSP]     = ps->leafmode? ERR_XOBJ : lls;
    SessnArray[RSA_SESS_CHECKSUM]  = chks;
    SessnArray[RSA_SESS_KM_REMOTE] = kmsrem;
    SessnArray[RSA_SESS_KM_LOCAL]  = kmsloc;
    s = xCreateSessn(rsaSessnInit, hlpRcv, hlpType, self, 4, SessnArray);

    if (!ps->leafmode) {
    	s->binding = mapBind(ps->actMap, &lls, s);
    	if ( s->binding == ERR_BIND ) {
	    xTrace0(rsasignp, TR_ERRORS, "rsaCreateSessn: mapBind error");
	    xClose(kmsrem);
 	    xClose(kmsloc);
	    xDestroy(s);
	    xFree((char *) ss);
	    return ERR_XOBJ;
    	}
    }

    xTrace1(rsasignp, TR_FULL_TRACE, "rsaCreateSessn: X_NEW(SState) returned %x", ss);

    ss->method = 0;
    ss->flags = 0;
    /* temporarily init method & rsakey here */
    ss->method = 2;
    ss->local_key = 0;
    ss->remote_key = 0;
    ss->chksize = xControl(chks, HASH_SIZE, 0, 0);
    ss->chksize = ROUND8(ss->chksize);

/* rsa_parse_key( */
    /* run together strings:  modulus, encryption exponent, decryption exponent */
/*
"13407807929942597099574024998205846127479365820592393377723561443721764030073217999499518551191939614890369525058823497765371118769307671100370276265584333"
" 6703903964971298549787012499102923063739682910296196688861780721860882015036773488400937149083451713845015929093243025426876941405973284973216824503042051"
" 2210719290388045346905339355787434075468743397471703259900205710040069873528157578187197280387011711560353471825220112394617831579775760441554445681825271");
*/
    s->state = (VOID *)ss;
    xTrace1(rsasignp, TR_MORE_EVENTS, "rsaCreateSessn returning %x",s);
    return s;
}


static xkern_return_t
rsaOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(rsasignp, TR_EVENTS, "rsaOpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);


    if ( defaultOpenEnable(ps->pasMap, hlpRcv, hlpType, 
	(VOID *)&hlpType) == XK_FAILURE ) {
        xTrace0(rsasignp, TR_MORE_EVENTS, "rsaOpenEnable defaultOpenEnable failed ");
        return XK_FAILURE;
    }
    
    val = xOpenEnable(self, hlpType, xGetDown(self,RSA_PROT_TRANSP), part);
    if ( val == XK_FAILURE ) {
        xTrace0(rsasignp, TR_MORE_EVENTS, "rsaOpenEnable xOpenEnable failed ");
        return XK_FAILURE;
    }
    

    xTrace1(rsasignp, TR_MORE_EVENTS, "rsaOpenEnable returning %x", val);
    return val;
} 


static xkern_return_t
rsaOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;



    xTrace4(rsasignp, TR_EVENTS, "rsaOpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    val = defaultOpenDisable(ps->pasMap, hlpRcv, hlpType, (void*)&hlpType);

    xTrace1(rsasignp, TR_MORE_EVENTS, "rsaOpenDisable returning %x", val);
    return val;
}

static xkern_return_t
rsaOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        s;
    PState      *pstate;
    Enable      *e;
    Part	part[2];
    char	partBuf[100];
   
    xTrace0(rsasignp, TR_MAJOR_EVENTS, "Entered rsaOpenDone");

    if ( self == hlpType ) {
        xTrace0(rsasignp, TR_ERRORS, "self == hlpType in rsaOpenDone");
        return XK_FAILURE;
    }
    pstate = (PState *)self->state;

    if (pstate->leafmode)
	return XK_FAILURE;
    /*
     * check for open enables
     */
    if ( mapResolve(pstate->pasMap, &hlpType, &e) == XK_FAILURE ) {
        /*
         * This shouldn't happen
         */
        xTrace0(rsasignp, TR_ERRORS,
                "rsaOpenDone: Couldn't find hlp for incoming session");
        return XK_FAILURE;
    }
    if (xControl(lls, GETPARTICIPANTS, partBuf, 100) <= 0) {
        xTrace0(rsasignp, TR_ERRORS, "rsaOpenDone: could not get participants");
        return XK_FAILURE;
    }
    partInternalize(part,partBuf);

    xDuplicate(lls);
    if ( (s = rsaCreateSessn(self, e->hlpRcv, e->hlpType, lls, part))
                == ERR_XOBJ ) {
        xClose(lls);
        return XK_FAILURE;
    }

    xTrace0(rsasignp, TR_EVENTS,
            "rsaOpenDone: Passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}



static xkern_return_t
rsaDemux( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  xkern_return_t val;
  
  xTrace3(rsasignp, TR_EVENTS, "rsaDemux entered: args %x %x %x",
	  self, lls, msg);

  xIfTrace(rsasignp, TR_DETAILED) msgShow(msg);

  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 
	xTrace0(rsasignp, TR_EVENTS, "rsaDemux: map resolve fails");
  	return XK_FAILURE;
  }

  if ( xIsSession(s) )  val = xPop(s, lls, msg, 0);

  if (val == XK_SUCCESS)
  xTrace0(rsasignp, TR_FUNCTIONAL_TRACE, "rsaDemux returning XK_SUCCESS");
  else if (val == XK_FAILURE)
  xTrace0(rsasignp, TR_FUNCTIONAL_TRACE, "rsaDemux returning XK_FAILURE");
  else
  xTrace1(rsasignp, TR_FUNCTIONAL_TRACE, "rsaDemux returning %x", val);

  return val;
}


/* $$$$ Erase rsakey! */
static xkern_return_t
rsaClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;
    xkern_return_t res;

    xTrace1(rsasignp, TR_MAJOR_EVENTS, "rsaClose: closing session %x", s);
    xAssert( xIsSession( s ) );
    xAssert( s->rcnt == 0 );
    if (!ps->leafmode)
    	res = mapRemoveBinding( ps->actMap, s->binding );
    xAssert( res != XK_FAILURE );
    if (!ps->leafmode)
	xClose(xGetDown(s,RSA_SESS_TRANSP));
    xClose(xGetDown(s,RSA_SESS_KM_REMOTE));
    xClose(xGetDown(s,RSA_SESS_KM_LOCAL));
    xDestroy( s );
    xTrace0(rsasignp, TR_DETAILED, "rsaClose returning");
    return XK_SUCCESS;
}


/* $$$$ Consider making rsacopy erase the input.
 This actually turns out to be a problem:
 There's no promise that we have sole ownership of the message data fragments:
 For example, the message may have been assembled with some constant pieces
 of boilerplate.  The crypt phase isn't free to erase those.
 This is a real problem, since the upper levels may recycle some fragments,
 and they really ought to be zero'd before being returned to the main pool.
 This also seems to prevent me from crypting in place, which I'd like as an
 option.
 It's less important, but less dangerous, to erase the incoming ciphertext
 after decryption, since I'm probably the sole owner.
 Further research is needed. */


static
long rsacopy(to,from,len,ignore)
    VOID *to; char *from; VOID *ignore; long len;
{
  xTrace4(rsasignp, TR_FULL_TRACE, "rsacopy args: %x %x %x %x",
	  to,from,len,ignore);

  bcopy((char *)from,(char *)to,len);
  return len;
}

static
void rsaStore(src, dst, len, arg)
    VOID *src;
    char *dst;
    long len;
    VOID *arg;
{
  xTrace4(rsasignp, TR_FULL_TRACE, "rsaStore args: %x %x %x %x",
	  src, dst, len, arg);

  bcopy((char *)src,(char *)dst,len);
}

static long msgpeek(to,from,len,ignore)
    void *to, *ignore; char *from; long len;
{   bcopy(from,(char *)to,len); return 0; }

static xmsg_handle_t
rsaPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
    long len, clen, i, chunk, copylen, j;
    unsigned long *ibuf, *obuf;
    MP_INT rsachunk, rsachunkout;
    char *cbuf,*cbuftmp,*pbuf,*pbuftmp;
    int crval, klval;
    u_char rsakey[2001];
    XObj kms = xGetDown(s, RSA_SESS_KM_LOCAL); /* use local key to sign */
    rsa_keyblock *key = ((SState *)s->state)->local_key; long keysizelong;
    char *chksum;
    int csize;
    SState *ss = (SState *)s->state;
    XObj chks = xGetDown(s, RSA_SESS_CHECKSUM);
    PState *ps = (PState *)(s->myprotl->state);

    xTrace2(rsasignp, TR_EVENTS, "rsaPush: session %x  msg %x", s, msg);

    klval = xControl(kms, KM_RESOLVE, (char *)rsakey, 2000);
    if (klval <= 0)
      {	/* We did not find a valid encryption key. */
	/* Don't send cleartext, so simply drop the message */
	xTrace0( rsasignp, TR_ERRORS,
		"ERROR: rsaPush() -- No encryption key found -- Dropping message");
	/* $$$$ maybe discard our existing key here? */
	return XK_FAILURE;
      }
    rsakey[2000] = 0;
    if (key)  /* if already have a key, see if it's changed */
    { if (strcmp((char *)rsakey,(char *)key->key_string)) rsa_zap_keyblock(key);
      else goto havekey; }
    key = rsa_parse_key(rsakey);  /* pretend success $$$ */
    ((SState *)s->state)->local_key = key;

  havekey:

    if (key->keysize < 2) return XK_FAILURE;
    xTrace2(rsasignp, TR_DETAILED, "rsaPush : pushing msg %x to checksum session %x"
	, msg, chks);
    /* find the checksum of the message */
    xPush(chks, msg);
    csize = ss->chksize;
    chksum = xMalloc(csize);
    msgPop(msg, chkPop, chksum, csize, 0);

    /* The input, plus a 4byte header for the length,
       is broken into chunks of keysize-1 bytes.
       Each chunk is encrypted, into a number of keysize bytes.
       The encrypted chunks are stuck together, making the output message. */
    len = csize;
    pbuf = xMalloc(len+sizeof(long));
    bcopy((char*)&len,pbuf,sizeof(long));
    chunk = (len + sizeof(long) + key->keysize - 2) / (key->keysize - 1);
    clen = chunk * key->keysize;
    cbuf = xMalloc(clen);
    keysizelong = (key->keysize + sizeof(long) - 1) / sizeof(long);
    ibuf = (unsigned long *) xMalloc(keysizelong * sizeof(long));  /* $$$$ wordsize dependency */
    obuf = (unsigned long *) xMalloc(keysizelong * sizeof(long));  /* $$$$ wordsize dependency */

    /* I discard the message attribute field on the way down.  "Feature." */
    bcopy(chksum, pbuf + sizeof(long), len); 

    /* $$$$ byte-order dependencies */
    for (i=0,pbuftmp=pbuf,cbuftmp=cbuf;i<chunk;i++,pbuftmp+=key->keysize-1,cbuftmp+=key->keysize)
    { ibuf[keysizelong-1] = 0;
      copylen = key->keysize-1;
      if (i == chunk-1) copylen = pbuf + len + sizeof(long) - pbuftmp;
      if (copylen <= 0 || copylen > key->keysize-1)
	{ printf("rsaPush: copylen bad %d\n",copylen); Kabort("rsaPush: copylen bad"); };
      bcopy(pbuftmp,(char *)ibuf,copylen);
      if (i == chunk-1 && copylen < key->keysize-1) bzero(((char *)ibuf)+copylen, key->keysize-1-copylen);
      /* kludge: setup number from ibuf, including 0 trimming */
      mpz_init(&rsachunk); mpz_init(&rsachunkout);
      for (j=keysizelong-1;j>=0;j--)
      { mpz_mul_2exp(&rsachunk,&rsachunk,32); mpz_add_ui(&rsachunk,&rsachunk,ibuf[j]); };
      /* compute chunk^dec (mod key) */
      mpz_powm(&rsachunkout,&rsachunk,&key->decexp,&key->modulus);
      /* kludge: get number into obuf */
      for (j=0;j<keysizelong;j++)
      { obuf[j] = mpz_get_ui(&rsachunkout);  mpz_div_2exp(&rsachunkout,&rsachunkout,32); };
      mpz_clear(&rsachunk); mpz_clear(&rsachunkout);
      bcopy((char *)obuf,cbuftmp,key->keysize); };
    xFree((char *)ibuf); xFree((char *)obuf);

/* $$$$ check that the prints refer to the correct buffers */

    xIfTrace(rsasignp, TR_FULL_TRACE)
    { printf("ptx "); for (i=0;i<len+sizeof(long) && i<40;i++) printf("%02x",0xff & *(pbuf+i)); printf("\n");
      if (len+sizeof(long)>=40) { printf("ptx ");
        for (i=(len+sizeof(long)<80)?40:(len+sizeof(long)-40);i<len+sizeof(long);i++) printf("%02x",0xff & *(pbuf+i)); printf("\n");};}
    xFree(pbuf);

    crval = 0;
    xIfTrace(rsasignp, TR_FULL_TRACE)
    { printf("ctx "); for (i=0;i<clen && i<40;i++) printf("%02x",0xff & *(cbuf+i)); printf("\n");
      if (clen>=40) { printf("ctx ");
        for (i=(clen<80)?40:(clen-40);i<clen;i++) printf("%02x",0xff & *(cbuf+i)); printf("\n");};}

    if (crval)
       xTrace1(rsasignp, TR_ERRORS, "rsaPush: Encryption barfed; error %D", crval);

    xFree(chksum);
    msgPush( msg, rsaStore, cbuf, clen, 0);

    if (!ps->leafmode && ps->sendsource) {
	char partBuf[100];

        if (xControl(xGetDown(s, RSA_SESS_TRANSP), GETPARTICIPANTS, partBuf, 100) <= 0) {
            xTrace0(rsasignp, TR_ERRORS, "rsaPush: could not get participants");
            return XK_FAILURE;
        }
	msgPush( msg, rsaStore, partBuf, 100, 0);
    }
    xTrace0(rsasignp, TR_DETAILED, "rsaPush exiting");
    if (ps->leafmode)
	return crval ? XK_FAILURE : XMSG_NULL_HANDLE;
    else {
    /* Discard message if encryption barfed */
    	if (!crval) return xPush(xGetDown(s, RSA_SESS_TRANSP), msg);  
    	return XK_FAILURE; 
    }
}

static long
chkPop(dst, src, len, arg)
    char *src; 
    void *dst;
    long len;
    void *arg;
{
    bcopy((char *)src, (char *)dst, len);
    return len;
}

static xkern_return_t
rsaPop( s, lls, m, hdr)
    XObj	s, lls;
    Msg 	*m;
    void        *hdr;
{
    long len,dlen; char* dbuf, *cbuf, *cbuftmp, *dbuftmp; int crval,i,j,chunk;
    unsigned long *ibuf, *obuf;
    u_char rsakey[2001];  MP_INT rsachunk, rsachunkout;
    XObj   kms = xGetDown(s,RSA_SESS_KM_REMOTE); /* use remote key to check signature */
    rsa_keyblock *key = ((SState *)s->state)->remote_key; long keysizelong;
    int klval;
    int clen;
    char *chksum, *signature;
    int csize;
    SState *ss = (SState *)s->state;
    XObj chks = xGetDown(s, RSA_SESS_CHECKSUM);
    PState *ps = (PState *)(s->myprotl->state);
    char message[200];

    xTrace4(rsasignp, TR_EVENTS, "rsaPop: args %x %x %x %x", s, lls, m, hdr);

    if (!ps->leafmode && ps->sendsource) {
	char sender[100];
	Part part[2];
	XObj myprotl = s->myprotl;
	Part partrem[1];
	XObj kmsrem;

    	msgPop(m, rsacopy, sender, 100, 0);  
	partInternalize (part, sender);

	/* open key manager session */
	partInit(partrem, 1);
	partrem[0] = part[RSA_PART_LOCAL];
	partrem[0].len =1; /* major hack */
	kmsrem = xOpen(myprotl, myprotl, xGetDown(myprotl, RSA_PROT_KMGR), partrem);
	if (kmsrem == ERR_XOBJ) {
	    xTrace0(rsasignp, TR_ERRORS, "rsaCreateSessn: could not open key manager");
	    xFree((char *) ss);
	    return XK_FAILURE;
	}
    	klval = xControl(kmsrem, KM_RESOLVE, (char *)rsakey, 2000);
	xClose(kmsrem);
    }
    else
    	klval = xControl(kms, KM_RESOLVE, (char *)rsakey, 2000);
    if (klval <= 0)
      {	/* We did not find a valid decryption key. */
	/* Can't decrypt, so simply drop the message */
	xTrace0( rsasignp, TR_ERRORS,
		"ERROR: rsaPop() -- No decryption key found -- Dropping message");
	/* $$$$ maybe discard our existing key here? */
	return XK_FAILURE;
      }
    rsakey[2000] = 0;
    if (key)  /* if already have a key, see if it's changed */
    { if (strcmp((char *)rsakey,(char *)key->key_string)) rsa_zap_keyblock(key);
      else goto havekey; }
    key = rsa_parse_key(rsakey);  /* pretend success $$$ */
    ((SState *)s->state)->remote_key = key;

  havekey:

    xIfTrace(rsasignp, TR_MAJOR_EVENTS)
    { printf("rsakey %x: %x %x %x %x %x %x %x %x   ", rsakey,
	   rsakey[0], rsakey[1], rsakey[2], rsakey[3],
	   rsakey[4], rsakey[5], rsakey[6], rsakey[7]);   }

    csize = ss->chksize;
    if (key->keysize < 2)
	return XK_FAILURE;
    chunk = (csize + sizeof(long) + key->keysize - 2) / (key->keysize - 1);
    clen = chunk * key->keysize;
    signature = xMalloc(clen+1);
    /* check the signature. First get the signature from the message */
    msgPop(m, rsacopy, signature, clen, 0);  

    xIfTrace (rsasignp, TR_MORE_EVENTS)
    { printf("signature: "); for(i=0;i<clen;i++) printf(" %08x",signature[i]); printf("\n"); };
    
    xTrace2(rsasignp, TR_DETAILED, "rsaPush : pushing msg %x to checksum session %x"
	, m, chks);
    /* find the checksum of the message */
    xPush(chks, m);
    chksum = xMalloc(csize+1);
    msgPop(m, chkPop, chksum, csize, 0);

    /* Decrypt here for now.  This should be broken out later. */
    len = clen;

    if (len <= 0)
    { xTrace1(rsasignp, TR_ERRORS,
	      "rsaPop received bad message, length %D must be positive",len);
      return XK_FAILURE; }
    if (len % key->keysize)
    { xTrace2(rsasignp, TR_ERRORS,
	      "rsaPop received bad message, length %D not a multiple of key size %D",len,key->keysize);
      return XK_FAILURE; }

    cbuf = xMalloc(len+1);
    bcopy(signature, cbuf, len);

    xIfTrace(rsasignp, TR_FULL_TRACE)
    { printf("ctx "); for (i=0;i<len && i<40;i++) printf("%02x",0xff & *(cbuf+i)); printf("\n");
      if (len>=40) { printf("ctx ");
        for (i=(len<80)?40:(len-40);i<len;i++) printf("%02x",0xff & *(cbuf+i)); printf("\n");};}

    /* decrypt */
    chunk = len / key->keysize;
    dbuf = xMalloc(len-chunk+1);  /* chunk * (keysize-1) = len - chunk */
    /* printf("rsaPop: dbuf = %x\n",dbuf); */
    keysizelong = (key->keysize + sizeof(long) - 1) / sizeof(long);
    ibuf = (unsigned long *) xMalloc(keysizelong * sizeof(long));  /* $$$$ wordsize dependency */
    obuf = (unsigned long *) xMalloc(keysizelong * sizeof(long));  /* $$$$ wordsize dependency */

    /* $$$$ byte-order dependencies */
    for (i=0,cbuftmp=cbuf,dbuftmp=dbuf;i<chunk;i++,cbuftmp+=key->keysize,dbuftmp+=key->keysize-1)
    { ibuf[keysizelong-1] = 0;
      bcopy(cbuftmp,(char *)ibuf,key->keysize);
      /* kludge: setup number from ibuf, including 0 trimming */
      mpz_init(&rsachunk); mpz_init(&rsachunkout); /*$$$$ dismal: calls malloc too often*/
      for (j=keysizelong-1;j>=0;j--)
      { mpz_mul_2exp(&rsachunk,&rsachunk,32); mpz_add_ui(&rsachunk,&rsachunk,ibuf[j]); };
      /* compute chunk^enc (mod key) */
      mpz_powm(&rsachunkout,&rsachunk,&key->encexp,&key->modulus);
      /* kludge: get number into obuf */
      for (j=0;j<keysizelong;j++)
      { obuf[j] = mpz_get_ui(&rsachunkout);  mpz_div_2exp(&rsachunkout,&rsachunkout,32); };
      mpz_clear(&rsachunk); mpz_clear(&rsachunkout);
/*$$$ should check that high byte is 0; and that unused bytes in last chunk are 0;
 and that length makes sense */
      bcopy((char *)obuf,dbuftmp,key->keysize-1); };
    xFree((char *)ibuf); xFree((char *)obuf); xFree(cbuf);
    crval = 0;

/* this shouldn't use len, but something else $$$$ */
    xIfTrace(rsasignp, TR_FULL_TRACE)
    { printf("ptx "); for (i=0;i<len && i<40;i++) printf("%02x",0xff & *(dbuf+i)); printf("\n");
      if (len>=40) { printf("ptx ");
        for (i=(len<80)?40:(len-40);i<len;i++) printf("%02x",0xff & *(dbuf+i)); printf("\n");};}

    if (crval)
    { xTrace1(rsasignp, TR_ERRORS, "rsaPop: Decryption barfed; error %D", crval);
      goto free; }

    dlen = *(long *)dbuf;
    xTrace1(rsasignp, TR_DETAILED, "rsaPop: dlen = %d", dlen);
    if (dlen<0 || dlen > chunk*(key->keysize-1)-4)  /* a start on the correct condition */
    { printf("rsaPop: recovered packet length bad, %D.  Original length %D.\n", dlen, len); crval = 1;
      goto free; }
/*
      Kabort("rsaPop: bad packet length"); };
*/


    /* This way, message attributes are preserved on the way up! */
    if (bcmp(chksum, dbuf + sizeof(long), csize) != 0) {
	/* comparison failed */
        crval = 1;
	goto free;
    }

    /* Fixup length in IP pseudoheader? */
    if (((SState *)(s->state))->flags & 1 || ((PState *)(xMyProtl(s)->state))->flags & 1)
    {  IPpseudoHdr *ipphdr = msgGetAttr( m, 0);
       if (ipphdr)
       {  ipphdr->len = htons(ntohs(ipphdr->len) - (len - dlen));
	  xTrace4( rsasignp, TR_FULL_TRACE,
   "rsaPop:  Adjusting IP pseudoheader length; ipphdr %x  phlen after %d  len %d  dlen %d",
		  ipphdr, ntohs(ipphdr->len), len, dlen);  }
       else { xTrace0( rsasignp, TR_FULL_TRACE,
	      "rsaPop:  Skipping length adjustment for IP pseudoheader - msg attr is 0."); };
     }

free:
    xFree(chksum);
    xFree(signature);
    xFree(dbuf);  /* $$$$ Should zero out cleartext. */
    msgPop (m, msgpeek, message, msgLen(m), 0);

    if (ps->leafmode)
	return crval ? XK_FAILURE : XK_SUCCESS;
    else 
    	return crval ? XK_FAILURE : xDemux(s, m);
} 


static int
rsaControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    PState *ps = (PState *)(s->myprotl->state);

    xAssert(xIsSession(s));
    switch (op) {

/* $$$$ need code */
      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;
	    if ((len < sizeof(max)) || (ps->leafmode)) return 0; 
	    if (xControl(xGetDown(s, RSA_SESS_TRANSP), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    *(int *)buf = max;
	    return(sizeof(max));
	}

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((SState *)(s->state))->flags |= 1;
	    xTrace0(rsasignp, TR_EVENTS,
	      "rsaControlSessn: Turning on IPpseudoheader length fixup.");  }
/* **** FALL THROUGH TO DEFAULT CASE, so lower sessions see it too.   **** */
/* **** Code for this control-op should be immediately above default: **** */

      default:
	if (ps->leafmode)
	    return -1;
        else if (buf && len) bzero(buf, len);   /* prevent info going down */
	    return xControl(xGetDown(s,RSA_SESS_TRANSP), op, buf, len);
    }
}


static int
rsaControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{
    PState *ps = (PState *)(self->state);

    xAssert(xIsProtocol(self));
    switch (op) {
      case GETMAXPACKET:
      case GETOPTPACKET:  /* $$$$ need code */
        {
            int max;
	    if ((len < sizeof(max)) || (ps->leafmode)) return 0; 
	    if (xControl(xGetDown(self, RSA_PROT_TRANSP), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
            *(int *)buf = max;
            return(sizeof(max));
	}

      /* for composablity rsa cannot zero buf */
      case ETH_REGISTER_ARP:
	{
	  return xControl(xGetDown(self, RSA_PROT_TRANSP), op, buf, len);
	}

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((PState *)(self->state))->flags |= 1;
	    xTrace0(rsasignp, TR_EVENTS,
	      "rsaControlProtl: Turning on IPpseudoheader length fixup.");  }
/* **** FALL THROUGH TO DEFAULT CASE, so lower protocols see it too.  **** */
/* **** Code for this control-op should be immediately above default: **** */

      default:
	if (ps->leafmode)
	    return -1;
	else {
          if (buf && len) bzero(buf, len);   /* prevent info going down */
	  return xControl(xGetDown(self, RSA_PROT_TRANSP), op, buf, len);
	}
  }
  
}


static void
rsaSessnInit(s)
    XObj s;
{
  s->push = rsaPush;
  s->pop = rsaPop;
  s->close = rsaClose;
  s->control = rsaControlSessn;
}

static xkern_return_t
romsendsource(self, ptr, fields, line, arg)
  XObj self;
  char **ptr;
  int  fields, line;
  void *arg;
{
  PState* ps = (PState*) self->state;
  xTrace0(rsasignp, TR_MAJOR_EVENTS, "rsasign sendsource rom opt");
  ps->sendsource = 1;

  return XK_SUCCESS;
}

